﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Buddy.Coroutines;
using Buddy.Wildstar.BotCommon;
using Buddy.Wildstar.Engine;
using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;

using log4net;



namespace Buddy.DefaultRoutine.Classes
{
    public abstract class BaseCombatClass
    {

        private static ILog Log = LogManager.GetLogger("BaseCombatClass");
        private static readonly PerFrameCachedValue<List<Player>> CachedPartyMembers = new PerFrameCachedValue<List<Player>>(() =>
        {
            // Cache the party member lookup. This really won't change often, so we could cache it for multiple frames.
            if (!GameManager.LocalPlayer.IsInGroup)
            {
                return new List<Player> { GameManager.LocalPlayer };
            }

            ulong groupId = GameManager.LocalPlayer.CurrentGroupId;
            return GameManager.Actors.Values.Where(a => a is Player && a.CurrentGroupId == groupId).Select(a => a as Player).ToList();
        });

        /// <summary>
        /// A list of party members within "view" of the player. If the player is not currently in a group, this will contain only the local player.
        /// </summary>
        protected List<Player> PartyMembers { get { return CachedPartyMembers; } }

        /// <summary>
        /// The local player
        /// </summary>
        protected Player Me { get { return GameManager.LocalPlayer; } }
        /// <summary>
        /// The local player's current target.
        /// </summary>
        protected Actor Target { get { return GameManager.LocalPlayer.CurrentTarget; } }

        internal virtual async Task<bool> Pull(Actor target)
        {
            return await Combat();
        }
        internal abstract Task<bool> Combat();

        internal virtual async Task<bool> Heal()
        {
            return false;
        }

        /// <summary>
        /// Tries to cast any of the supplied abilities (in order). Returns true if any of them were cast.
        /// </summary>
        /// <param name="spellNames"></param>
        /// <returns></returns>
        public async Task<bool> CastAny(params string[] spellNames)
        {
            foreach (string spellName in spellNames)
            {
                if (await Cast(spellName))
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Tries to cast the specified ability on yourself.
        /// </summary>
        /// <param name="spell"></param>
        /// <returns></returns>
        public async Task<bool> SelfCast(string spell)
        {
            return await CastOn(spell, () => Me);
        }

        public bool HasSpell(string spellName)
        {
            return SpellManager.GetSpell(spellName) != null;
        }




        public async Task<bool> PopFlask()
        {
            //if (!Me.Buffs.Exists(a => a != null && a.Name.Contains("Flask of"))) //Keeps chucking out null reference exceptions.
			if (!(Me.HasBuff("Flask of Speed") || Me.HasBuff("Flask of Reputation") || Me.HasBuff("Flask of Greed") || Me.HasBuff("Flask of Harvesting") || Me.HasBuff("Flask of Prestige") || Me.HasBuff("Flask of Renown") || Me.HasBuff("Flask of Experience"))) //Forcing me to do this the lame way
            {
                var flaskSlot = SpellManager.AvailableSpells.FirstOrDefault(a => a != null && !string.IsNullOrEmpty(a.Name) && a.Name.Contains("Flask") && a.CanCast);
                if (flaskSlot != null)
                {
                    flaskSlot.Cast();
                    await Coroutine.Yield();
                    return true;
                }
            }
            return false;
        }


        public async Task<bool> ChugPotion()
        {
            // Chug a potion. Note, this only chugs Instant Healing and Sustained Healing
            var potionSlot = SpellManager.GetMiscSpell(MiscSpellType.UsePotion);
            if (potionSlot != null && potionSlot.Name.Contains("Healing") && Me.HealthPercent < 50 && potionSlot.CanCast)
            {
                potionSlot.Cast();
                await Coroutine.Yield();
                return true;
            }
            return false;
        }








        /// <summary>
        /// Returns true if the specified spell is equal, or higher than the tier specified.
        /// </summary>
        /// <param name="spellName"></param>
        /// <param name="tier"></param>
        /// <returns></returns>
        public bool IsSpellTier(string spellName, int tier)
        {
            SpellManager.SpellInformation spell = SpellManager.GetSpell(spellName);
            return spell != null && spell.Tier >= tier;
        }

        /// <summary>
        /// Returns true if the specified spell has been proced.
        /// </summary>
        /// <param name="spellName"></param>
        /// <returns></returns>
        public bool IsSpellProcReady(string spellName)
        {
            SpellManager.SpellInformation spell = SpellManager.GetSpell(spellName);
            return spell != null && spell.IsSpellReady && spell.CanCast;
        }

        /// <summary>
        /// Returns true if the specified spell is being casted.
        /// </summary>
        /// <param name="spellName"></param>
        /// <returns></returns>
        public bool CastingSpell(string spellName)
        {
            SpellManager.SpellInformation spell = SpellManager.GetSpell(spellName);
            return Me.CurrentCast != null && spell != null && Me.CurrentCast.BaseSpellId == spell.Ability.BaseSpellId;
        }

        /// <summary>
        /// Returns true if the specified spell's cooldown is less than or equal to the given argument, including 0 being off-cooldown
        /// </summary>
        /// <param name="spellName"></param>
        /// <param name="timeRemaining"></param>
        /// <returns></returns>
        public bool spellCooldown(string spellName, double timeRemaining)
        {
            SpellManager.SpellInformation spell = SpellManager.GetSpell(spellName);
            return spell != null && timeRemaining >= spell.CooldownTimeRemaining.TotalSeconds;
        }


        /// <summary>
        /// Returns true if the given AMP is enabled in the user's current LAS
        /// </summary>
        /// <param name="AMPname"></param>
        /// <returns></returns>
        public static bool HasAMP(string AugName)
        {
            return GameManager.Lua.GetReturnValue<bool>(@"for _,AMP in pairs(AbilityBook.GetEldanAugmentationData(AbilityBook.GetCurrentSpec()).tAugments) do
		if (AMP.strTitle == '" + AugName + @"') then
			return (AMP.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated)
			end
		end
		return false");
        }







        /// <summary>
        /// Attempts to cast the specified spell. Returns true if successful.
        /// </summary>
        /// <param name="spell"></param>
        /// <param name="onSelf"></param>
        /// <returns></returns>
        public async Task<bool> Cast(string spell, bool onSelf = false)
        {
            if (onSelf)
            {
                return await CastOn(spell, () => Me);
            }
            return await CastOn(spell, () => Me.CurrentTargetGuid != null ? Me.CurrentTarget : Targeting.BestTarget);
        }

        /// <summary>
        /// Casts the specified spell on the specified actor. Returns false if the spell is not known, on cooldown, not ready, or you don't have the available resources to cast it.
        /// </summary>
        /// <param name="spellName">The name of the spell to be cast.</param>
        /// <param name="on">A function returning the actor to have the spell cast on. If this is not the current target, it will be targeted.</param>
        /// <param name="ignoreMoveInRange">Whether to ignore moving in range for the current ability.</param>
        /// <param name="ignoreFacing">Whether we should ignore automatically facing the target when casting, and during casting of the current ability. (Useful for non-targeted abilities)</param>
        /// <param name="charges">If the ability is a "charge and release" ability, how many charges to build up before releasing it.</param>
        /// <param name="cancellation">A function returning true if the current cast should be aborted, and control given back to the caller.</param>
        /// <returns></returns>
        public async Task<bool> CastOn(string spellName, Func<Actor> on, bool ignoreMoveInRange = false, bool ignoreFacing = false, int charges = 3, Func<bool> cancellation = null)
        {
            // TODO: For ignoreMoveInRange and ignoreFacing, since we set client arguments, we may need to "tweak" them during this cast so that the client doesn't automatically face
            // us towards things, or move into range.


            SpellManager.SpellInformation spell = SpellManager.GetSpell(spellName);

            // Do we know the spell? If not, don't try anything further.
            if (spell == null)
            {
                return false;
            }

            // Make sure we're close to the target for this ability
            // And facing it.
            // We use the client to get within range, but this is just going to be to ensure that we get a "path" to the object correct
            // May want to exclude it to be honest
            if (on() != Me)
            {
                if (on() != Target)
                {
                    // We need to make sure that we wait for it to be targeted.
                    // TODO: Maybe make this use a Coroutine.Wait(()=>Me.CurrentTarget == on()) ?
                    on().Target();
                    await Coroutine.Sleep(100);
                }

                // This is instant.
                if (!ignoreFacing)
                {
					if (Target != null)
						{on().Face();}
                }
                else
                {
                    Me.Target();
                }
                if (Target != null && !ignoreMoveInRange && spell.TargetMaxRange > 0 && on().Distance > spell.TargetMaxRange)
                {
                    Log.Debug("Moving within " + spell.TargetMaxRange + " of " + on() + " to cast " + spell.Name);
                    var pos = on().Position;
                    if (await CommonBehaviors.MoveWithin(() => pos, spell.TargetMaxRange, true, cancellation: () => cancellation != null && cancellation() || @on() == null) ==
                        CommonBehaviors.BehaviorResult.Cancelled)
                        return true;
                }

            }

            // NOTE: Almost all abilities will fail to cast if we're sprinting.
            // So make sure that we're not. Thanks!
            if (GameManager.Movement.IsSprinting)
                GameManager.Movement.Sprint(true);

            // Force the bot to stop any movement.
            if (GameManager.LocalPlayer.IsMoving)
                Navigator.Stop();

            if (!spell.CanCast)
            {
                return false;
            }

            DefaultRoutine.Log.Info("Casting " + spellName);

            if (spell.Ability.IsGroundTargeted)
            {
                // If the ability is ground-targeted
                // Then we want to press the ability once to get it on the mouse
                // Then set the position of the mouse to wherever (we use the target by default)
                spell.Cast();
                GameManager.CurrentSpellPlacementPosition = on().Position;
            }

            switch (spell.SpellCastMethod)
            {
                // In the case of auras, it's almost always "instant"
                // So just cast it, and continue on
                // TODO: Double-check that I'm actually correct here...
                case CastMethod.Aura:
                    spell.Cast();
                    break;

                // Cast time -> cast goes off
                case CastMethod.Normal:
                // Channeled just means that the cast happens as long as we're casting (instead of finishing at the end)
                case CastMethod.Channeled:
                case CastMethod.ChanneledField:
                // Multiphase means that there are areas of the ability that do more damage. In almost all cases, this is the "middle"
                // So ensure we're facing
                case CastMethod.Multiphase:
                    // Cast the ability, and wait for it to finish casting, or the cancellation to cause it to stop
                    spell.Cast();
                    if (await Coroutine.Wait(500, () => Me.IsCasting || (cancellation != null && cancellation())))
                    {
                        await Coroutine.Wait(TimeSpan.FromMilliseconds(200), () => !Me.IsCasting || (cancellation != null && cancellation()));
                    }
                    break;
                case CastMethod.PressHold:
                    // Button down
                    spell.Cast(true);

                    // Wait for it to start casting...
                    if (await Coroutine.Wait(500, () => Me.IsCasting || (cancellation != null && cancellation())))
                    {
                        while (Me.IsCasting)
                        {
                            if (!ignoreFacing)
                            {
                                on().Face();
                            }
                            await Coroutine.Yield();
                        }
                    }

                    // Wait for it to finish casting (cast time + 200ms) or the cancellation to stop it
                    await Coroutine.Wait(spell.CastTime.Add(TimeSpan.FromMilliseconds(200)), () => !Me.IsCasting || (cancellation != null && cancellation()));

                    // Button up
                    spell.Cast(false);

                    break;
                case CastMethod.RapidTap:
                    spell.Cast();

                    // CurrentThreshold is 1 when there are charges left on rapid tap abilities
                    while (spell.CurrentThreshold != 0)
                    {
                        if (cancellation != null && cancellation())
                        {
                            return false;
                        }
                        if (!ignoreFacing)
                        {
                            on().Face();
                        }

                        spell.Cast();
                        await Coroutine.Sleep(100);
                    }

                    break;
                case CastMethod.ChargeRelease:	//Working decently enough, further improvements will need to be made to allow the use of off-gcd abilities while Charging however.

                    //Need to ensure Multi Tap is on when using Charged Abilities, the core logic tries turning it off every time the bot is started.
                    if (GameManager.Lua.GetReturnValue<bool>("return Apollo.GetConsoleVariable('spell.useButtonDownForAbilities')") == true)
                    {
                        GameManager.Lua.DoString("Apollo.SetConsoleVariable('spell.useButtonDownForAbilities', false)");
                    }



                    int ChargeTime = 0;

                    spell.Cast();
                    await Coroutine.Sleep(500); //Too low and the bot will get stuck in the cast

                    if (GameManager.LocalPlayer.CurrentCast != null)
                    {
                        ChargeTime = (int)(GameManager.Lua.GetReturnValue<float>(@"return GameLib.GetSpell(" + GameManager.LocalPlayer.CurrentCast.SpellId + @"):GetThresholdTime()") * 1000) - 400;
                        if (ChargeTime < 0)
                        { ChargeTime = 0; }
                    }

                    if (await Coroutine.Wait(500, () => Me.IsCasting))
                    {

                        await Coroutine.Wait(ChargeTime, () => !Me.IsCasting); //Ideally we need a check to see if the ability is fully charged,
                        //	await Coroutine.Wait(10000, () => spell.CurrentThreshold >= charges);
                    }

                    spell.Cast();

                    break;

                // I'm not sure what these are to be completely honest...
                case CastMethod.Transactional:
                case CastMethod.Unused04:
                case CastMethod.ClientSideInteraction:
                    // I guess just cast it?
                    spell.Cast();
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }

            return true;
        }
    }
}