﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Buddy.Coroutines;
using Buddy.DefaultRoutine.Classes;
using Buddy.Wildstar.BotCommon;
using Buddy.Wildstar.Engine;
using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;

using log4net;

namespace Buddy.DefaultRoutine
{
    public class DefaultRoutine : CombatRoutineBase, IUIButtonProvider
    {
	    internal static ILog Log = LogManager.GetLogger("Default Routine");
		private BaseCombatClass _currentClass;

	    private BaseCombatClass CurrentClass
	    {
		    get
		    {
			    if (_currentClass == null)
			    {
				    switch (GameManager.LocalPlayer.Class)
				    {
					    case Class.Medic:
						    _currentClass = new Medic();
						    break;
						case Class.Engineer:
							_currentClass = new Engineer();
						    break;
						case Class.Stalker:
						    _currentClass = new Stalker();
						    break;
						case Class.Warrior:
						    _currentClass = new Warrior();
						    break;
						case Class.Spellslinger:
						    _currentClass = new SpellSlinger();
						    break;
						case Class.ESPer:
							_currentClass = new Esper();
						    break;
				    }
			    }
			    return _currentClass;
		    }
	    }

	    /// <summary>
	    /// Called when this routine is selected for use.
	    /// </summary>
	    public override void OnSelected()
		{
			using (GameManager.Memory.AcquireFrame(true))
		    {
			    // player.ignoreAlwaysFaceTarget = set to false, the player will face the target when casting abilities
			    // player.disableFacingLock = set to false, the player will constantly face the target during the casting of abilities (coincides with the above, except it's done constantly, rather that at cast time)
			    // player.moveToTargetOnSelfAOE = set to true, the player will move within range of the target when casting abilities (ensures the game puts us within range of the target when casting)
			    GameManager.Lua.DoString(
				    "Apollo.SetConsoleVariable('player.ignoreAlwaysFaceTarget', false) " +
				    "Apollo.SetConsoleVariable('player.disableFacingLock', false) " +
				    "Apollo.SetConsoleVariable('player.moveToTargetOnSelfAOE', true) ");

			    // This then ensures that we're using the correct casting model for the logic we use.
			    // Button presses cause events, rather than overlapping "down" states.
			    SpellManager.SetSpellCastingBehavior(SpellCastingBehavior.Normal);
		    }
		    base.OnSelected();
	    }

		#region Overrides of CombatRoutineBase

	    /// <summary>
	    ///     The name of this authored object.
	    /// </summary>
	    public override string Name { get { return "Default Combat Routine"; } }

	    /// <summary>
	    ///     The author of this object.
	    /// </summary>
	    public override string Author { get { return "Bossland GmbH"; } }

	    /// <summary>
	    ///     The version of this object implementation.
	    /// </summary>
	    public override Version Version { get { return new Version("1.0"); } }

	    public override async Task Combat()
	    {
			// So basically, if we don't have a target, or it's dead
			// OR it happens to be something we're not in combat with (common during quests where we interact with something, and doing so spawns mobs)
			// Set a new target.
		    if (GameManager.LocalPlayer.CurrentTargetGuid == 0 || GameManager.LocalPlayer.CurrentTarget.IsDead || !Targeting.CombatTargets.Contains(GameManager.LocalPlayer.CurrentTarget))
		    {
			    var newTarget = Targeting.BestTarget;
			    if (newTarget != null)
			    {
				    newTarget.Target();
				    await Coroutine.Yield();
			    }
			    else
			    {
				    return;
			    }
		    }

		    await CurrentClass.Combat();
	    }

	    public override async Task Pull(object preferredTarget)
	    {
			// Safe-cast to Actor. (Null if not an actor, otherwise, an Actor object)
		    Targeting.PullTarget = preferredTarget as Actor;

		    if (Targeting.PullTarget == null)
		    {
			    Log.Error("Non-Actor object passed into Pull method. Don't know what to do!");
			    return;
		    }

			// Ensure it's targeted
		    if (GameManager.LocalPlayer.CurrentTargetGuid != Targeting.PullTarget.Guid)
		    {
			    Targeting.PullTarget.Target();
			    await Coroutine.Sleep(125);
		    }


		    await CurrentClass.Pull(Targeting.PullTarget);
	    }

	    #endregion

	    #region Implementation of IUIButtonProvider

	    /// <summary>
	    /// The text to be displayed on the button in the UI
	    /// </summary> 
	    public string ButtonText
	    {
		    get
		    {
			    // NOTE: this doesn't actually show anything. TODO: Set it on the tooltip for the settings button
			    return "Default Routine Test Button";
		    }
	    }

	    /// <summary>
	    /// Called when the button is clicked.
	    /// </summary>
	    /// <param name="sender"></param>
	    public void OnButtonClicked(object sender)
	    {
			Log.Info(GameManager.InputManager.FindKeybindForAction(InputAction.MoveForward).ToString());
	    }

	    #endregion

		/// <summary>
		/// Instructs the bot to wait for a specified amount of time. Useful for resting, or other things that need the bot to "wait" before carrying on other logic.
		/// Note: This does not stop in-combat quest behavior mechanics.
		/// </summary>
		/// <param name="time"></param>
	    public void PauseBotFor(TimeSpan time)
		{
			// This is all it takes to tell the bot to "wait" for us to do something.
			GameEngine.BotPulsator.Suspend(GameEngine.CurrentBot, time);
		}
    }
}
